package fr.mrdev023.vulkan_java.vk.utils;

import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.lwjgl.system.MemoryStack;
import org.lwjgl.vulkan.VK10;
import org.lwjgl.vulkan.VkLayerProperties;
import org.tinylog.Logger;

public class InstanceValidationLayersSelector {
    private boolean withValidationLayers = false;

    public InstanceValidationLayersSelector withValidationLayers(boolean withValidationLayers) {
        this.withValidationLayers = withValidationLayers;
        return this;
    }

    public List<String> selectValidationLayers() {
        if (!withValidationLayers) {
            return Collections.emptyList();
        }

        var supportedLayers = getSupportedValidationLayers();
        log("Supported validation layers", supportedLayers);

        var layersToUse = selectValidationLayersFromSupportedLayers(supportedLayers);
        log("Selected validation layers", layersToUse);

        if (layersToUse.isEmpty() && withValidationLayers) {
            Logger.warn("Request validation but no supported validation layers found. Falling back to no validation");
        }

        return layersToUse;
    }

    private List<String> getSupportedValidationLayers() {
        try (MemoryStack stack = MemoryStack.stackPush()) {
            // Get the number of supported validation layers by the current vulkan instance
            IntBuffer numLayersArr = stack.callocInt(1);
            VK10.vkEnumerateInstanceLayerProperties(numLayersArr, null);
            int numLayers = numLayersArr.get(0);

            // Get the supported validation layers by the current vulkan instance
            VkLayerProperties.Buffer propsBuf = VkLayerProperties.calloc(numLayers, stack);
            VK10.vkEnumerateInstanceLayerProperties(numLayersArr, propsBuf);

            // Convert the supported validation layers to a list of strings
            List<String> supportedLayers = new ArrayList<>();
            for (int i = 0; i < numLayers; i++) {
                VkLayerProperties props = propsBuf.get(i);
                String layerName = props.layerNameString();
                supportedLayers.add(layerName);
            }

            return supportedLayers;
        }
    }

    private List<String> selectValidationLayersFromSupportedLayers(List<String> supportedLayers) {
        List<String> layersToUse = new ArrayList<>();

        // Main validation layer
        if (supportedLayers.contains("VK_LAYER_KHRONOS_validation")) {
            layersToUse.add("VK_LAYER_KHRONOS_validation");
            return layersToUse;
        }

        // Fallback 1
        if (supportedLayers.contains("VK_LAYER_LUNARG_standard_validation")) {
            layersToUse.add("VK_LAYER_LUNARG_standard_validation");
            return layersToUse;
        }

        // Fallback 2 (set)
        List<String> requestedLayers = new ArrayList<>();
        requestedLayers.add("VK_LAYER_GOOGLE_threading");
        requestedLayers.add("VK_LAYER_LUNARG_parameter_validation");
        requestedLayers.add("VK_LAYER_LUNARG_object_tracker");
        requestedLayers.add("VK_LAYER_LUNARG_core_validation");
        requestedLayers.add("VK_LAYER_GOOGLE_unique_objects");

        return requestedLayers.stream().filter(supportedLayers::contains).toList();
    }

    private void log(String title, List<String> layers) {
        Logger.debug("{} ({})", title, layers.size());
        for (String layer : layers) {
            Logger.debug(" - {}", layer);
        }
    }
}